from math import cos
from openseneca.interfaces.providers import BaseProvider
from openseneca.interfaces.response import ProviderResponse
from openseneca.utils.logger import Logger
from openseneca.costs import CostCalculator
from typing import Generator
import http.client
import logging
import requests
import time
import json

# ----------------------------
# TODO, make this configurable.
# Enable debug output in the http.client module
# http.client.HTTPConnection.debuglevel = 1
# Create a http logger
# logging.basicConfig()
# logging.getLogger().setLevel(logging.DEBUG)
# requests_log = logging.getLogger("requests.packages.urllib3")
# requests_log.setLevel(logging.DEBUG)
# requests_log.propagate = True
# ----------------------------


logger = Logger()

class AzureProvider(BaseProvider):
  def __init__(self, stream: bool = False):
    """
    Initializes an instance of the Azure class.

    Args:
      stream (bool, optional): Whether to enable streaming. Defaults to False.
    """
    self.stream = stream

  def request(self,
         endpoint: str,
         authentication_header: dict = None,
         body: dict = None,
         content_type: str = 'application/json',
         ) -> Generator[ProviderResponse, None, None]:
    """
    Sends a request to the specified endpoint using the specified authentication
    header and body.

    Args:
      endpoint (str): The URL endpoint to send the request to.
      authentication_header (dict, optional): The authentication header
        to include in the request. Defaults to None.
      body (dict, optional): The request body. Defaults to None.
      content_type (str, optional): The content type of the request.
        Defaults to 'application/json'.

    Returns:
      Generator: A generator that yields ProviderResponse objects.

    """
    if body is None:
      body = {}

    if authentication_header is None:
      authentication_header = {}

    headers = {
      'Content-Type': content_type
    }
    stream = self.stream

    headers.update(authentication_header)
    logger.info(f"Is streaming: {stream}")

    if stream:
      body.update({
        "stream": stream
      })

    retry_attempts = 3
    retry_delay = 5
    costs_calulator = CostCalculator(self.llm_name)

    for attempt in range(retry_attempts):
      logger.debug(f"Sending request to {endpoint}")
      logger.debug(f"Body: {body}")
      logger.debug(f"Attempt {attempt + 1} of {retry_attempts}")
      try:

        response = requests.post(
          endpoint,
          json=body,
          headers=headers,
          timeout=120,  # TODO: Make this configurable
          stream=stream
        )

        if stream:
          json_object = {}
          # Process the response in chunks
          for chunk in response.iter_lines():
            if chunk.startswith(b"data:"):
              json_data = chunk.split(b"data:", 1)[1].strip()
              if len(json_data) > 0:
                try:
                  json_object = json.loads(json_data)
                  if 'choices' in json_object \
                      and 'finish_reason' in json_object['choices'][0] \
                      and json_object['choices'][0]['finish_reason']:
                    if json_object and 'usage' in json_object:
                      _usage = json_object['usage']
                      logger.info(f"Usage: {_usage}")

                      costs = costs_calulator.calculate(
                        _usage['prompt_tokens'],
                        _usage['completion_tokens']
                      )
                      logger.info(f"Costs: {costs}")
                      yield ProviderResponse(
                        200, {}, json.dumps(json_object), cost=costs
                      )
                      break
                  yield ProviderResponse(200, {}, json.dumps(json_object))
                except json.JSONDecodeError:
                  pass  # silence, it's an empty response.
            logger.debug(f"Chunk: {chunk}")
          time.sleep(0.1)

        else:  # not streaming, just return the entire response
          logger.info('not streaming')
          json_object = json.loads(response.text)
          if json_object and 'usage' in json_object:
            _usage = json_object['usage']
            logger.info(f"Usage: {_usage}")

            costs = costs_calulator.calculate(
              _usage['prompt_tokens'],
              _usage['completion_tokens']
            )

            final_response = ProviderResponse(
              response.status_code,
              response.headers,
              response.text,
              cost=costs
            )

          else:
            # mh?
            final_response = ProviderResponse(
              response.status_code,
              response.headers,
              response.text,
            )

          logger.debug(f"Response: {final_response}")
          if 'error' in json.loads(final_response.content):
            logger.error(f"Error in response: {final_response.content}")
            yield ProviderResponse(0, {}, "{}")

          yield final_response
      except requests.exceptions.Timeout:
        if attempt < retry_attempts - 1:
          time.sleep(retry_delay)
        else:
          yield ProviderResponse(0, {}, "{}")

      return


class Categories:

  @staticmethod
  def get():
    return {
      'text_greetings': 0,
      'language_translation': 1,
      'creative_writing': 2,
      'data_analysis': 3,
      'code_python': 4,
      'math_equation': 5,
      'code_generation': 6,
      'code_javascript': 7,
      'text_poetry': 8,
      'nonfiction_research': 9,
      'customer_service': 10,
      'technology_exploration': 11,
      'history_search': 12,
      'technology_analysis': 13,
      'philosophical_question': 14,
      'code_java': 15,
      'code_c++': 16,
      'technology_tutorial': 17,
      'technology_support': 18,
      'computer_programming': 19,
      'software_development': 20,
      'code_c#': 21,
      'history_analysis': 22,
      'health_advice': 23,
      'content_creation': 24,
      'psychology_support': 25,
      'math_problem': 26,
      'business_analysis': 27,
      'linguistic_analysis': 28,
      'technology_review': 29,
      'place_search': 30,
      'text_analysis': 31,
      'science_study': 32,
      'scientific_explanation': 33,
      'fact_checking': 34,
      'science_exploration': 35,
      'historical_analysis': 36,
      'international_law': 37,
      'science_theory': 38,
      'science_solution': 39,
      'code_php': 40,
      'finance_analysis': 41,
      'finance_advice': 42,
      'psychology_theory': 43,
      'science_tutorial': 44,
      'code_sql': 45,
      'science_introduction': 46,
      'business_management': 47,
      'place_analysis': 48,
      'education_tutorial': 49,
      'business_creation': 50,
      'biographies_of_figures': 51,
      'travel_advice': 52,
      'web_design': 53,
      'business_support': 54,
      'math_analysis': 55,
      'customer_engagement': 56,
      'translation_teaching': 57,
      'philosophy_theory': 58,
      'marketing_support': 59,
      'philosophy_question': 60,
      'gaming_analysis': 61,
      'philosophy_analysis': 62,
      'philosophy_exploration': 63
    }


import yaml
from openseneca.utils.logger import Logger
import os
import pkg_resources
logger = Logger()


class CostCalculator:
  def __init__(self, llm_name):
    self.llm_name = llm_name
    self.prompt_tokens_cost = None
    self.completion_tokens_cost = None
    self.load_costs_from_config()

  def load_costs_from_config(self):
    local_path = os.path.dirname(os.path.abspath(__file__))
    config_file = f'{local_path}/../config.yml'
    if not os.path.exists(config_file):
      config_file = pkg_resources.resource_filename('openseneca', 'config.yml')

    with open(f'{local_path}/config.yml', 'r') as config_file:
      config = yaml.safe_load(config_file)
      llm_config = config.get('models', {}).get(self.llm_name, {})
      if not llm_config:
          llm_config = config.get('models', {}).get('_DEFAULT', {})

      llm_config = config.get('models', {}).get('_DEFAULT', {})
      logger.debug(f"LLM config: {llm_config}")
      self.prompt_tokens_cost = llm_config.get('costs', {}).get('prompt_tokens', 0)
      self.completion_tokens_cost = llm_config.get('costs', {}).get('completion_tokens', 0)
      logger.debug(f"Prompt tokens cost: {self.prompt_tokens_cost}")
      logger.debug(f"Completion tokens cost: {self.completion_tokens_cost}")

  def calculate(self, prompt_tokens, completion_tokens):
    total_cost = (prompt_tokens * self.prompt_tokens_cost/1000) +\
      (completion_tokens * self.completion_tokens_cost/1000)
    return float(total_cost)


import requests
from openseneca.utils.logger import Logger
import json

logger = Logger()

class ProviderResponse:

  cost = 0.00  # default cost

  def __init__(self,
      status_code: int = 200,
      headers: dict = None,
      content: str = '{"status": "error", "message": "No content"}',
      cost: float = 0
    ):

    if headers is None:
      headers = {}

    self.status_code = status_code
    self.headers = headers
    self.content = content

    if cost:
      if cost:
        content_json = json.loads(self.content)
        content_json["usage"]["cost"] = cost
        self.content = json.dumps(content_json)

  @staticmethod
  def from_requests_response(response: requests.Response):
    return ProviderResponse(
      response.status_code,
      response.headers,
      response.text
    )


  def print_properties(self):
    logger.info(f"status_code: {self.status_code}")
    logger.info(f"headers: {self.headers}")
    logger.info(f"content: {self.content}")
    logger.info(f"cost: {self.cost}")



import inspect
import importlib
import os
from openseneca.interfaces.models import ModelBuilder
from openseneca.utils.logger import Logger

logger = Logger()


class ModelBuilderNotFound(Exception):
  """
  Exception raised when the model or its builder is not found.
  """
  pass


def get_model_builder(name) -> ModelBuilder:
  """
  Retrieves the model builder class for a given model name.

  Args:
    name (str): The name of the model.

  Returns:
    ModelBuilder: The model builder class corresponding to the given model name.

  Raises:
    ModelBuilderNotFound: If the model builder class is not found for
      the given model name.
  """
  module = importlib.import_module('openseneca.models')
  for _, cls in inspect.getmembers(module, inspect.isclass):
    if hasattr(cls, 'NAME'):
      print(getattr(cls, 'NAME'))
    if hasattr(cls, 'NAME') and getattr(cls, 'NAME') == name:
      builder_class_name = cls.__name__.replace('Model', 'Builder')
      builder_class = getattr(module, builder_class_name, None)
      if builder_class is not None:
        return builder_class
  err = f"Model builder not found for model {name}"
  logger.error(err)
  raise ModelBuilderNotFound(err)


def get_LLMs():
  """
  Retrieves the LLMs (Logical Link Monitors) from the environment variables and
    returns them sorted alphabetically.

  Returns:
    list: A sorted list of LLMs (Logical Link Monitors).
  """
  os_vars = {
    key: value
    for key, value in os.environ.items()
    if key.startswith("oS__")
  }

  LLMs = set()
  for key, value in os_vars.items():
    if key.endswith("_ENDPOINT") or key.endswith("_AUTH"):
      name = key.split("__")[1].replace("_ENDPOINT", "").replace("_AUTH", "")
      LLMs.add(name)

  # Sort the LLMs by name (A-Z)
  LLMs = sorted(LLMs)

  return LLMs



from venv import logger
from torch import nn
import torch
from transformers import BertTokenizer, BertModel
import numpy as np
import time
from huggingface_hub import hf_hub_download
from openseneca.utils.logger import Logger
from openseneca.classify.categories import Categories
import os

from dotenv import load_dotenv
load_dotenv()

logger = Logger()

class OpenSenecaLLM(nn.Module):

  MAX_SEQ = 128

  def __init__(self):
      super(OpenSenecaLLM, self).__init__()
      self.bert = BertModel.from_pretrained(
        'google-bert/bert-base-multilingual-cased'
      )
      self.dropout = nn.Dropout(0.0)
      self.linear = nn.Linear(768, len(Categories.get()))
      self.relu = nn.ReLU()

      current_path = os.path.dirname(os.path.abspath(__file__))
      self.path = f'{current_path}/openseneca-llm-v01/openseneca-llm-v01.pt'

      if not os.path.exists(self.path):
        try:
          # Download the model
          hf_hub_download(
            repo_id="OpenSeneca/openseneca-llm-v01",
            filename="openseneca-llm-v01.pt",
            local_dir=f"{current_path}/openseneca-llm-v01/"
          )
        except Exception as e:
          logger.error(f"Error downloading model: {e}")
          raise Exception("Error downloading the OpenSeneca-LLM.")

      self.__tokenizer = \
        BertTokenizer.from_pretrained(
          'google-bert/bert-base-multilingual-cased'
        )
      self.labels = Categories.get()
      self.__load()

  def forward(self, input_id, mask):

    _, pooled_output = self.bert(input_ids= input_id, attention_mask=mask,return_dict=False)
    dropout_output = self.dropout(pooled_output)
    linear_output = self.linear(dropout_output)
    final_layer = self.relu(linear_output)

    return final_layer

  def __load(self):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    self.load_state_dict(torch.load(self.path, map_location=device))
    self.eval()

  def get_label(self, text: str) -> str:
    start_time = time.time()

    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    encoded_input = self.__tokenizer(
      text,
      max_length=self.MAX_SEQ,
      padding='max_length',
      truncation=True,
      return_tensors='pt',
    )

    mask = encoded_input['attention_mask'].to(device)
    input_id = encoded_input['input_ids'].squeeze(1).to(device)
    output = self(input_id, mask).cpu().detach().numpy()
    predictions = np.argmax(output, axis=1)
    label_id = predictions[0]

    label = self.__search_label(self.labels, label_id)
    end_time = time.time()
    execution_time_ms = (end_time - start_time) * 1000

    logger.debug("OpenSenecaLLM Execution time: {} milliseconds."\
      .format(execution_time_ms))

    return label

  @staticmethod
  def __search_label(labels, label_id):
    for package_name, package_label_id in labels.items():
        if package_label_id == label_id:
            return package_name
    return None



import logging
import colorlog
import inspect
import os
import threading

class ThreadIDFilter(logging.Filter):
    """
    This is a filter which injects the thread id into the log.

    This because the Logger class is being instantiated in the main thread,
    so the thread ID (thid) is being set at the time of instantiation and
    doesn't change afterwards. This means that all log messages, even those
    from other threads, will have the same thread ID.

    To print the correct thread ID for each log message, you can use a
    logging.Filter to add the thread ID to the log record at the time the
    log message is processed.

    e.g: handler.addFilter(ThreadIDFilter())

    """
    def filter(self, record):
        record.threadid = threading.current_thread().ident
        return True

class Logger:
    """
    A Logger class for colored logging.
    """

    def __init__(self, name: str = None, level: str = 'DEBUG'):
        """
        Initialize the logger with a name and level.
        If no name is provided, it defaults to the name of the module where the Logger is instantiated.
        """

        if name is None:
            frame = inspect.stack()[1]
            module = inspect.getmodule(frame[0])
            name = os.path.split(module.__file__)[-1]

        pid = os.getpid()
        thid = threading.current_thread().ident
        formatter = colorlog.ColoredFormatter(
            f"%(log_color)s%(asctime)s | 🚦 {pid}/%(threadid)s | 📁 \033[90m{name} %(log_color)s| "\
            "%(levelname)-4s | \033[37m%(message)s%(reset)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            reset=True,
            log_colors={
                'DEBUG':    'cyan',
                'INFO':     'green',
                'WARNING':  'yellow',
                'ERROR':    'red',
                'CRITICAL': 'red,bg_white',
            },
            style='%'
        )

        self.logger = logging.getLogger(name)
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        handler.addFilter(ThreadIDFilter())
        self.logger.addHandler(handler)
        self.logger.setLevel(getattr(logging, level))

    def debug(self, message: str):
        """Log a debug message."""
        self.logger.debug(message)

    def info(self, message: str):
        """Log an info message."""
        self.logger.info(message)

    def warning(self, message: str):
        """Log a warning message."""
        self.logger.warning(message)

    def error(self, message: str):
        """Log an error message."""
        self.logger.error(message)

    def critical(self, message: str):
        """Log a critical message."""
        self.logger.critical(message)


from .gpt3 import (
  GPT35ChatBuilder,
  GPT35ChatModel
)
from .gpt4 import (
  GPT4ChatBuilder,
  GPT4ChatModel
)
from .llama2 import (
  LLama27BChatBuilder,
  LLama213BChatBuilder,
  LLama270BChatBuilder,
  LLama27BChatModel,
  LLama213BChatModel,
  LLama270BChatModel,
)

from .llama3 import (
  LLama38BChatBuilder,
  LLama370BChatBuilder,
  LLama38BChatModel,
  LLama370BChatModel,
)

from .cohere import (
  CohereCommandRChatBuilder,
  CohereCommandRPlusChatBuilder,
  CohereCommandRChatModel,
  CohereCommandRPlusChatModel,
)

from .mistral import (
  MistralLargeChatBuilder,
  MistralLargeChatModel
)


from __future__ import annotations
from abc import ABC, abstractmethod
from openseneca.interfaces.response import ProviderResponse
from typing import Generator

class BaseProvider(ABC):
  """
  Base class for providers (Azure, HF, Google, etc.)
  """

  llm_name = None

  @abstractmethod
  def request(self,
        endpoint: str,
        authentication_header: dict = None,
        body: dict = None,
        content_type: str = 'application/json',
        ) -> Generator[ProviderResponse, None, None]:
    """
    Abstract method for making a request.

    Returns:
      Generator: A generator that yields ProviderResponse objects.
    """
    return ProviderResponse()

  def set_llm_name(self, llm_name: str):
    """
    Sets the name of the LLM.

    Args:
      llm_name (str): The name of the LLM.
    """
    self.llm_name = llm_name


from torch import nn


class StaticOpenSenecaLLMInstance:
  ollm = None

  @staticmethod
  def set_instance(instance: nn.Module):
    """
    Sets the instance of the OpenSenecaLLM model.

    Args:
      instance (nn.Module): The instance of the OpenSenecaLLM model to be set.

    """
    if StaticOpenSenecaLLMInstance.ollm is None:
      StaticOpenSenecaLLMInstance.ollm = instance



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Generator
from openseneca.interfaces.providers import BaseProvider
import os
from typing import List, Dict, Any
from dotenv import load_dotenv
import pkg_resources
from openseneca.interfaces.response import ProviderResponse
load_dotenv()
import yaml

def import_model_settings(model_name) -> dict:
    local_path = os.path.dirname(os.path.abspath(__file__))
    config_file = f'{local_path}/../config.yml'
    if not os.path.exists(config_file):
      config_file = pkg_resources.resource_filename('openseneca', 'config.yml')
    with open(config_file, 'r') as file:
        config = yaml.safe_load(file)
        if model_name not in config["models"]:
          return config["models"]["_DEFAULT"]['settings']
        return config["models"][model_name]['settings']

class ModelSetupNotFound(Exception):
  """
  Exception raised when the endpoint or bearer token is not found
  during model setup.
  """
  pass

class ModelBuilder(ABC):
  """
  The Model Creator Class is the interface that declares the factory method
  to return a Model Class (e.g. Llama2Model).
  """

  def __init__(
      self,
      provider: BaseProvider,
      temperature: float = 0.0,
      top_p: float = 0.0,
      max_tokens: int = 0,
    ):
    self.provider = provider
    self.temperature = temperature
    self.top_p = top_p
    self.max_tokens = max_tokens

  @abstractmethod
  def build(self) -> Model:
    """
    This method have to be implemented by the concrete creators, build and
    return the Model Class.
    """
    pass


class Model(ABC):
  """
  Represents an AI model.
  """

  NAME = 'MODEL'  # it should be overwritten by the concrete model.

  def __init__(
      self,
      provider: BaseProvider,
      temperature: float = 0.0,
      top_p: float = 0.0,
      max_tokens: int = 0,
    ):
    """
    Initializes a new instance of the Model class.

    Args:
      provider (BaseProvider): The provider object.
      endpoint (str): The endpoint URL.
      bearer_token (str): The bearer token for authentication.

    Raises:
      ModelSetupNotFound: If the endpoint or the bearer_token is empty.

    """
    self.provider = provider
    self.provider.set_llm_name(self.NAME)
    self.endpoint = os.getenv(f'oS__{self.NAME}_ENDPOINT', None)
    self.auth_token = os.getenv(f'oS__{self.NAME}_AUTH', None)
    self.settings = import_model_settings(self.NAME)

    if temperature:
      self.settings['temperature'] = temperature

    if max_tokens:
      self.settings['max_tokens'] = max_tokens

    if top_p:
      self.settings['top_p'] = top_p

    if not self.endpoint or not self.auth_token:
        raise ModelSetupNotFound(
          f"{self.NAME} endpoint or authentication token not found" \
            " in the .env file."
          )

class ChatModel(Model):
  """
  The Product interface declares the operations that all concrete products
  must implement.
  """

  @abstractmethod
  def request(self,
        messages: List[Dict[Any, Any]],
        ) -> Generator[ProviderResponse, None, None]:
    """
    Sends a request with the given messages and returns a response.

    Args:
      messages (List[Dict[Any, Any]]): A list of messages to be sent.

    Returns:
      Generator[ProviderResponse, None, None]: A generator that yields
      ProviderResponse objects.
    """
    pass



from typing import List, Dict, Any
from openseneca.interfaces.models import (
  ModelBuilder,
  ChatModel,
)


class GPT35ChatBuilder(ModelBuilder):
  """
  The Builder can do some complex initialization here or transform data
  if needed before passing it to the Model.

  https://learn.microsoft.com/en-us/azure/ai-services/openai/reference
  """


  def build(self) -> ChatModel:
    return GPT35ChatModel(
      provider=self.provider,
      temperature=self.temperature,
      top_p=self.top_p,
      max_tokens=self.max_tokens,
    )


class GPT35ChatModel(ChatModel):
  """
  Represents a chat model for OpenAI GPT3.X.
  """

  NAME = 'GPT3.5_CHAT_TURBO'

  def request(self,
      messages: List[Dict[Any, Any]],
    ) -> str:

    body = {
      "messages": messages,
    }

    body.update(self.settings)

    response = self.provider.request(
      endpoint=self.endpoint,
      authentication_header={'api-key': self.auth_token},
      body=body
    )
    return response


from setuptools import setup, find_packages

# test: pip install -e .
# rm -rf dist/ && python setup.py sdist bdist_wheel
# twine upload --repository openseneca dist/*

VERSION = '0.0.6'
DESCRIPTION = 'OpenSeneca'
LONG_DESCRIPTION = ''

def read_requirements():
    with open('openseneca/requirements.txt', 'r') as req:
        content = req.read()
        requirements = content.split('\n')

    return requirements

# Setting up
setup(
        name="openseneca",
        version=VERSION,
        author="Ottavio Fogliata",
        author_email="ottavio.fogliata@openseneca.ai",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=read_requirements(),
        package_data={
          'openseneca': ['router.pk', 'config.yml'],
        },
        keywords=['python']
)


from typing import List, Dict, Any
from openseneca.interfaces.models import (
  ModelBuilder,
  ChatModel,
)


class LLama27BChatBuilder(ModelBuilder):
  """
  The Builder can do some complex initialization here or transform data
  if needed before passing it to the Model.

  https://learn.microsoft.com/en-gb/azure/ai-studio/how-to/deploy-models-llama?tabs=azure-studio
  """

  def build(self) -> ChatModel:
    return LLama27BChatModel(provider=self.provider)


class LLama27BChatModel(ChatModel):
  """
  Represents a chat model for Llama2.

  Inherits from the ChatModel class.
  """

  NAME = 'LLAMA2_7B_CHAT'

  def request(self,
      messages: List[Dict[Any, Any]],
    ) -> str:

    body = {
      "messages": messages,
      "stop": [
        "### Human",
        "### Assistant",
        "<|eot_id|>",
        "<|start_header_id|>"
      ]
    }

    body.update(self.settings)

    response = self.provider.request(
      endpoint=self.endpoint,
      authentication_header={'Authorization': f'Bearer {self.auth_token}'},
      body=body
    )
    return response


class LLama213BChatBuilder(ModelBuilder):
  """
  The Builder can do some complex initialization here or transform data
  if needed before passing it to the Model.

  https://learn.microsoft.com/en-gb/azure/ai-studio/how-to/deploy-models-llama?tabs=azure-studio
  """

  def build(self) -> ChatModel:
    return LLama213BChatModel(provider=self.provider)


class LLama213BChatModel(ChatModel):
  """
  Represents a chat model for Llama2.

  Inherits from the ChatModel class.
  """

  NAME = 'LLAMA2_13B_CHAT'

  def request(self,
      messages: List[Dict[Any, Any]],
    ) -> str:

    body = {
      "messages": messages,
      "stop": [
        "### Human",
        "### Assistant",
        "<|eot_id|>",
        "<|start_header_id|>"
      ]
    }

    body.update(self.settings)

    response = self.provider.request(
      endpoint=self.endpoint,
      authentication_header={'Authorization': f'Bearer {self.auth_token}'},
      body=body
    )
    return response


class LLama270BChatBuilder(ModelBuilder):
  """
  The Builder can do some complex initialization here or transform data
  if needed before passing it to the Model.

  https://learn.microsoft.com/en-gb/azure/ai-studio/how-to/deploy-models-llama?tabs=azure-studio
  """

  def build(self) -> ChatModel:
    return LLama270BChatModel(
      provider=self.provider,
      temperature=self.temperature,
      top_p=self.top_p,
      max_tokens=self.max_tokens,
    )


class LLama270BChatModel(ChatModel):
  """
  Represents a chat model for Llama2.

  Inherits from the ChatModel class.
  """

  NAME = 'LLAMA2_70B_CHAT'

  def request(self,
      messages: List[Dict[Any, Any]],
    ) -> str:

    body = {
      "messages": messages,
      "stop": [
        "### Human",
        "### Assistant",
        "<|eot_id|>",
        "<|start_header_id|>"
      ]
    }

    body.update(self.settings)

    response = self.provider.request(
      endpoint=self.endpoint,
      authentication_header={'Authorization': f'Bearer {self.auth_token}'},
      body=body
    )
    return response



import pandas as pd
from openseneca.classify.llm import OpenSenecaLLM
from openseneca.classify.static import StaticOpenSenecaLLMInstance
from langdetect import detect
from openseneca.interfaces.models import Model
from openseneca.utils.inspection import get_model_builder, get_LLMs
from openseneca.utils.logger import Logger
from typing import List, Dict, Any
import pkg_resources

import os
from dotenv import load_dotenv
import os
import time
load_dotenv()
logger = Logger()


class Router:

  default_model = "GPT3.5_CHAT_TURBO"

  # TODO: cheapest model preference.

  def __init__(self, source: str):
    start_time = time.time()
    self.dataframe = pd.read_pickle(source)

    llms = get_LLMs()
    logger.info("LLMS configured: " + str(llms))
    # Filter the dataframe to only include the models in the LLMs list.
    self.dataframe = self.dataframe[self.dataframe['model_name'].isin(llms)]

    end_time = time.time()
    logger.debug("Time to load the router file: "\
      f"{round((end_time - start_time) * 1000, 2)} ms.")

  def get_best(
    self,
    prompt,
  ):
    """
    Returns the best model name based on the given criteria.

    Args:
      category (str): The category of the model.
      language (str): The language of the model.
      context (bool): Whether the model requires context.
      instruction (bool): Whether the model requires instruction.
      is_chat (bool): Whether the model is for chat purposes.

    Returns:
      str: The name of the best model that matches the given criteria.
    """
    filtered_df = self.dataframe[
      (self.dataframe['category_name'] == prompt.label) &
      (self.dataframe['language'] == prompt.language) &
      (self.dataframe['context'] == prompt.is_context_based) &
      (self.dataframe['instruction'] == prompt.is_instructed) &
      (self.dataframe['is_chat'] == prompt.is_chat)
    ]

    try:
      filtered_df = filtered_df.sort_values(by='elo_score', ascending=False)
      best_model = filtered_df.iloc[0]['model_name']
      temperature = filtered_df.iloc[0]['temperature']
      top_p = filtered_df.iloc[0]['top_p']
    except (ValueError, IndexError):
      return self.__get_best_by_language(
        prompt.label,
        prompt.language,
        prompt.is_chat
      )

    return best_model, temperature, top_p

  def __get_best_by_language(
    self,
    category: str,
    language: str,
    is_chat: bool = True,
  ):
    filtered_df = self.dataframe[
        (self.dataframe['category_name'] == category) &
        (self.dataframe['language'] == language) &
        (self.dataframe['is_chat'] == is_chat)
    ]

    try:
      filtered_df = filtered_df.sort_values(by='elo_score', ascending=False)
      best_model = filtered_df.iloc[0]['model_name']
      temperature = filtered_df.iloc[0]['temperature']
      top_p = filtered_df.iloc[0]['top_p']

    except (ValueError, IndexError):
      return self.default_model, 0.7, 0.8
      # return self.__get_best_category_only(category, is_chat)

    return best_model, temperature, top_p

  # def __get_best_category_only(
  #   self,
  #   category: str,
  #   is_chat: bool = True,
  # ):
  #   filtered_df = self.dataframe[
  #       (self.dataframe['category_name'] == category) &
  #       (self.dataframe['is_chat'] == is_chat)
  #   ]

  #   try:
  #     filtered_df = filtered_df.sort_values(by='elo_score', ascending=False)
  #     best_model = filtered_df.iloc[0]['model_name']
  #     temperature = filtered_df.iloc[0]['temperature']
  #     top_p = filtered_df.iloc[0]['top_p']
  #   except (ValueError, IndexError):
  #     return self.default_model, 0.5, 0.92

  #   return best_model, temperature, top_p


router_path = os.path.dirname(os.path.abspath(__file__)) + '/router.pk'
if not os.path.exists(router_path):
  router_pk_path = pkg_resources.resource_filename('openseneca', 'router.pk')
router = Router(router_path)

class Prompt:
  messages: List[Dict[Any, Any]]
  user_prompt: str
  language: str
  label: str
  # TODO: context, instruction, is_chat must be set from api params
  #   or autodetected (!?)
  is_context_based: bool = False # TODO
  is_instructed: bool = False  # TODO
  is_chat: bool = True
  best_llm_builder: Model = None
  temp = 0.8
  top_p = 0.9

  def __init__(self, messages: List[Dict[Any, Any]]):

    StaticOpenSenecaLLMInstance\
      .set_instance(OpenSenecaLLM())

    self.messages = messages

    if messages[-1]['role'] != 'user':
      raise Exception("Last message is not from user")

    self.user_prompt = messages[-1]['content']
    self.language = detect(self.user_prompt)
    self.label = StaticOpenSenecaLLMInstance.ollm.get_label(self.user_prompt)
    logger.info(f"Detected language: {self.language}")
    logger.info(f"Detected label: {self.label}")

  def classify(self,):
      best_llm, temp, top_p = router.get_best(self)
      logger.info(f"Best LLM: {best_llm, temp, top_p}")
      best_llm_builder = get_model_builder(best_llm)
      logger.info(f"Builder: {best_llm_builder}")
      self.best_llm_builder = best_llm_builder
      self.temp = temp
      self.top_p = top_p
      return self
